/* sim.c: simulator functions
 * $Id: sim.c,v 1.1 2001/09/05 10:50:34 martin Exp $
 */

/* This file is part of `exhaust', a memory array redcode simulator.
 * Copyright (C) 2000 M Joonas Pihlaja
 *    
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * You may contact the author by:
 *   email: jpihlaja@@cc.helsinki.fi
 *   mail:  M Joonas Pihlaja
 *          Artturi Kannistontie 4A11
 *          00320, Helsinki
 *          Finland
 */

/*
 * Thanks go to the pMARS authors and Ken Espiritu whose ideas have
 * been used in this simulator.  Especially Ken's effective addressing
 * calculation code in pMARS 0.8.6 has been adapted for use here.
 */

#include <stdio.h>
#include <stdlib.h>

#include "exhaust.h"
#include "asm.h"
#include "sim.h"
#include "insn.h"

/* DEBUG level:
 *     0: none
 *   >=1: disassemble each instruction (no output)
 *     2: print each instruction as it is executed
 */

/*
 * File scoped stuff
 */

/* internal warrior structure */
typedef struct w_st {
  field_t *tail;        /* next free location to queue a process */
  field_t *head;        /* next process to run from queue */
  struct w_st *succ;        /* next warrior alive */
  u32_t nprocs;         /* number of live processes in this warrior */
  struct w_st *pred;        /* previous warrior alive */
  int id;           /* index (or identity) of warrior */
} w_t;


#define DEF_MAX_WARS 2
#define DEF_CORESIZE 8000
#define DEF_PROCESSES 8000
#define DEF_CYCLES 80000

static
unsigned int Coresize=0,
             Processes=0;
static int NWarriors = 0, Cycles = 0;


static w_t *War_Tab = NULL;
static insn_t *Core_Mem = NULL;
static field_t *Queue_Mem = NULL;

/* protos */
static int sim_proper( int, field_t *, int * );





/* NAME
 *     sim_alloc_bufs, sim_free_bufs -- alloc and free buffers of sim_proper()
 * 
 * SYNOPSIS
 *     insn_t *sim_alloc_bufs( int nwars, unsigned int coresize,
 *                             unsigned int processes, unsigned int cycles );
 *     void sim_free_bufs();
 * 
 * INPUTS
 *     nwar        -- number of warriors
 *     coresize    -- size of core
 *     processes   -- max no of processes / warrior
 *     cycles      -- the number of cycles to play before tie.
 * 
 * RESULTS
 *     These functions manage the core, queue, and w_t warrior info
 *     struct array memories.
 *
 *     Core_Mem, Queue_Mem, and War_Tab memories are allocated
 *     or freed as requested.  Any earlier memories are freed before
 *     new allocations are made.
 *
 * RETURN VALUE
 *     sim_alloc_bufs(): the address of core memory, or NULL if
 *           out of memory.
 *     sim_free_bufs(): none
 *
 * GLOBALS
 *     All file scoped globals 
 */

void
sim_free_bufs()
{
  if ( Core_Mem ) free( Core_Mem ); Core_Mem = NULL; Coresize = 0;
  if ( Queue_Mem ) free( Queue_Mem ); Queue_Mem = NULL; Processes = 0;
  if ( War_Tab ) free( War_Tab ); War_Tab = NULL; NWarriors = 0;
}


insn_t *
sim_alloc_bufs( nwars, coresize, processes, cycles )
     int nwars;
     unsigned int coresize, processes, cycles;
{
  unsigned int queue_size, old_queue_size;
  unsigned int old_nwars;

  /* allocate Core buffer */
  if ( Coresize != coresize ) {
    if ( Core_Mem ) free( Core_Mem ); Coresize = 0;
    if (( Core_Mem = malloc( sizeof(insn_t) * coresize ) ))
      Coresize = coresize;
  }

  /* allocate Queue buffer */
  old_queue_size = NWarriors*Processes+1;
  queue_size = nwars*processes+1;
  old_nwars = NWarriors;
  if ( queue_size != old_queue_size ) {
    if ( Queue_Mem ) free( Queue_Mem ); NWarriors = Processes = 0;
    if (( Queue_Mem = malloc( sizeof(field_t)*queue_size ) )) {
      NWarriors = nwars;
      Processes = processes;
    }
  }

  /* allocate War_Tab array */
  if ( nwars != (int)old_nwars ) { /*** (int) added by Martin Ankerl 2000 **/
    if ( War_Tab ) free( War_Tab ); NWarriors = 0;
    if (( War_Tab = malloc( sizeof(w_t)*nwars ) )) NWarriors = nwars;
  }

  Cycles = cycles;
  if ( Core_Mem && Queue_Mem && War_Tab ) return Core_Mem;
  sim_free_bufs();
  return NULL;
}


/* NAME
 *     sim, sim_mw -- public functions to simulate a round of Core War
 * 
 * SYNOPSIS
 *     int sim_mw( int nwar, field_t *war_pos_tab, int *death_tab );
 *     int sim( int nwar, field_t w1_start, field_t w2_start,
 *      int cycles_arg, void **ptr_result );
 * 
 * INPUTS
 *     nwar        -- number of warriors
 *     w1_start, w2_start -- core addresses of first processes
 *                    warrior 1 and warrior 2. Warrior 1 executes first.
 *     cycles      -- the number of cycles to play before tie.
 *     ptr_result  -- NULL, except when requesting for the address of core.
 *     war_pos_tab -- core addresses where warriors are loaded in
 *            the order they are to be executed.
 *     death_tab   -- the table where dead warrior indices are stored
 * 
 * DESCRIPTION
 *     The real simulator is inside sim_proper() to which sim() and
 *     sim_mw() are proxies.  sim_mw() reads the warrior position
 *     of the ith warrior from war_tab_pos[i-1].
 *
 * RESULTS
 *     The warriors fight their fight in core which gets messed up in
 *     the process.  For sim(): If nwar == -1 then a core and process
 *     queues of default size for max. two warriors are allocated and
 *     the address of the core memory is returned via the ptr_result
 *     pointer.
 *
 *     sim_mw() stores indices of warriors that die into the death_tab
 *     array in the order of death.  Warrior indices start from 0.
 * 
 * RETURN VALUE
 *     sim_mw(): the number of warriors still alive at the end of the
 *       battle.
 *       -1: simulator panic attack -- something's gone wrong
 *
 *     sim(): 
 *       single warrior: 0: warrior suicided, 1: warrior didn't die.
 *       one-on-one two warriors:
 *        0: warrior 1 won, warrior 2 won, 2: tie
 *       -1: simulator panic attack -- something's gone wrong
 * 
 * GLOBALS
 *     All file scoped globals
 */

int
sim_mw( nwar, war_pos_tab, death_tab )
     int nwar;
     field_t *war_pos_tab;
     int *death_tab;
{
  if ( !Core_Mem || !Queue_Mem || !War_Tab ) return -1;
  return sim_proper( nwar, war_pos_tab, death_tab );
}


int
sim( int nwar,
     field_t w1_start,
     field_t w2_start,
     int cycles,
     void **ptr_result )
{
  field_t war_pos_tab[2];
  int death_tab[2];
  int alive_cnt;

  /* if the caller requests for the address of core, allocate 
   * the default buffers and give it
   */
  if ( nwar < 0 ) {
    if ( nwar == -1 && ptr_result ) {
      *ptr_result = sim_alloc_bufs( DEF_MAX_WARS, DEF_CORESIZE,
                    DEF_PROCESSES, DEF_CYCLES );
      return 0;
    }
    return -1;
  }
  if ( nwar > 2 ) return -1;

  /* otherwise set up things for sim_mw() */
  Cycles = cycles;
  war_pos_tab[0] = w1_start;
  war_pos_tab[1] = w2_start;

  alive_cnt = sim_mw( nwar, war_pos_tab, death_tab );
  if ( alive_cnt < 0 ) return -1;

  if ( alive_cnt == 2 ) return 2; /* tie */
  return death_tab[0] ? 0 : 1;
}


/*-------------------------------------------------------------------------
 * private functions
 */

/* NAME
 *     sim_proper -- the real simulator code
 * 
 * SYNOPSIS
 *     int sim_proper( int nwar, field_t *war_pos_tab, int *death_tab );
 * 
 * INPUTS
 *     nwar        -- number of warriors
 *     war_pos_tab -- core addresses where warriors are loaded in
 *            the order they are to be executed.
 *     death_tab   -- the table where dead warrior indices are stored
 * 
 * RESULTS
 *     The warriors fight their fight in core which gets messed up in
 *     the process.  The indices of warriors that die are stored into
 *     the death_tab[] array in the order of death.  Warrior indices
 *     start from 0.
 * 
 * RETURN VALUE
 *     The number of warriors still alive at the end of the
 *     battle or -1 on an anomalous condition.
 * 
 * GLOBALS
 *     All file scoped globals
 */

/* Various macros:
 *
 *  queue(x): Inserts a core address 'x' to the head of the current
 *            warrior's process queue.  Assumes the warrior's
 *        tail pointer is inside the queue buffer.
 *
 * x, y must be in 0..coresize-1 for the following macros:
 *
 * INCMOD(x): x = x+1 mod coresize
 * DECMOD(x): x = x-1 mod coresize
 * ADDMOD(z,x,y): z = x+y mod coresize
 * SUBMOD(z,x,y): z = x-y mod coresize
 */

#define queue(x)  do { *w->tail++ = (x); if ( w->tail == queue_end )\
                                          w->tail = queue_start; } while (0)

#define INCMOD(x) do { if ( ++(x) == coresize ) (x) = 0; } while (0)
#define DECMOD(x) do { if ((x)-- == 0) (x) = coresize1; } while (0)
#define ADDMOD(z,x,y) do { (z) = (x)+(y); if ((z)>=coresize) (z) -= coresize; } while (0)
#define SUBMOD(z,x,y) do { (z) = (x)-(y); if ((int)(z)<0) (z) += coresize; } while (0)


int
sim_proper( nwar, war_pos_tab, death_tab )
     int nwar;
     field_t *war_pos_tab;
     int *death_tab;
{
    /*
     * Core and Process queue memories.
     *
     * The warriors share a common cyclic buffer for use as a process
     * queue which the contains core addresses where active processes
     * are.  The buffer has size N*P+1, where N = number of warriors,
     * P = maximum number of processes / warrior.
     *
     * Each warrior has a fixed slice of the buffer for its own process
     * queue which are initially allocated to the warriors in reverse
     * order. i.e. if the are N warriors w1, w2, ..., wN, the slice for
     * wN is 0..P-1, w{N-1} has P..2P-1, until w1 has (N-1)P..NP-1.
     *
     * The core address of the instruction is fetched from the head of
     * the process queue and processes are pushed to the tail, so the
     * individual slices slide along at one location per executed
     * instruction.  The extra '+1' in the buffer size is to have free
     * space to slide the slices along.
     * 
     * For two warriors w1, w2:
     *
     * |\......../|\......../| |
     * | w2 queue | w1 queue | |
     * 0          P         2P 2P+1
     */

    insn_t *core;
    field_t *queue_start, *queue_end; /* queue mem. start, end */

    /*
     * Cache Registers.
     *
     * The '94 draft specifies that the redcode processor model be
     * 'in-register'.  That is, the current instruction and the
     * instructions at the effective addresses (ea's) be cached in
     * registers during instruction execution, rather than have
     * core memory accessed directly when the operands are needed.  This
     * causes differences from the 'in-memory' model.  e.g. MOV 0,>0
     * doesn't change the instruction's b-field since the instruction at
     * the a-field's effective address (i.e. the instruction itself) was
     * cached before the post-increment happened.
     *
     * There are conceptually three registers: IN, A, and B.  IN is the
     * current instruction, and A, B are the ones at the a- and
     * b-fields' effective addresses respectively.
     *
     * We don't actually cache the complete instructions, but rather
     * only the *values* of their a- and b-field.  This is because
     * currently there is no way effective address computations can
     * modify the opcode, modifier, or addressing modes of an
     * instruction.  Also, we reuse the registers when there's no overlap
     * in lifetimes (makes less of a hassle for the register allocator).
     */

    u32_t in;         /* 'in' field of current insn for decoding */
    u32_t ra_a, ra_b,     /* A register values */
        rb_a, rb_b;     /* B register values */
    u32_t da, db;         /* effective address of instruction's
			   * a- and b-fields */
    /* 
     * alias some register names to real variables
     */
#define in_a ra_a
#define in_b rb_b

    field_t *pofs;
    insn_t *pt;
    unsigned int mode;

    /*
     * misc.
     */
    w_t *w;           /* current warrior */
    u32_t ip;         /* current instruction pointer */
    u32_t ftmp, t;        /* temps */
    unsigned int coresize, coresize1; /* size of core, size of core - 1 */
    int cycles;           /* instructions to execute before tie counter */
    int alive_cnt;

#if DEBUG >= 1
    insn_t insn;          /* used for disassembly */
    char debug_line[256];     /* ditto */
#endif



    /*
     * Setup queue and core pointers
     */
    alive_cnt = nwar;
    cycles = nwar * Cycles; /* set instruction executions until tie counter */
    coresize = Coresize;
    coresize1 = Coresize-1;
    core = Core_Mem;
    queue_start = Queue_Mem;
    queue_end = Queue_Mem + NWarriors*Processes+1;

    /* Setup War_Tab and links all around */
    pofs = queue_end-1;       /* init. wars[] table */
    War_Tab[0].succ = &War_Tab[nwar-1];
    War_Tab[nwar-1].pred = &War_Tab[0];
    t = nwar-1;
    ftmp = 0;
    do {
	if ( t > 0 ) War_Tab[t].succ = &War_Tab[t-1];
	if ( (int)t < (nwar-1) ) War_Tab[t].pred = &War_Tab[t+1]; /*** (int) added by Martin Ankerl 2000 **/
	pofs -= Processes;
	*pofs = war_pos_tab[ftmp];
	War_Tab[t].head = pofs;
	War_Tab[t].tail = pofs+1;
	War_Tab[t].nprocs = 1;
	War_Tab[t].id = ftmp;
	--t;
	ftmp++;
    } while ( (int)ftmp < nwar );


    /*
     * Main loop is run for each executed instruction
     */
    w = &War_Tab[ nwar-1 ];
    do {

	ip = *w->head++;        /* load current instruction */
	if ( w->head == queue_end ) w->head = queue_start;
	in = iMASK & core[ ip ].in; /* iMASK & clears all flags */
	in_a = core[ ip ].a;
	in_b = core[ ip ].b;

#if DEBUG >= 1
	insn = core[ip];
	dis1( debug_line, insn, coresize);
#endif

	/*
	 * a-field effective address calculation.
	 *
	 * This bit was assimilated from Ken Espiritu's 
	 * pmars 0.8.6 patch.  Ditto for the b-field 
	 * calculation.
	 */
	mode = in & mMASK;
	in >>= mBITS;       /* shift out a-mode bits */
	rb_a = in_a;        /* necessary if B-mode is immediate */

	if ( mode != IMMEDIATE ) {
	    ADDMOD( da, in_a, ip );
	    pt = core+da;     /* pt is where field points to (without */
	    /* indirection). */

	    if ( mode != DIRECT ) {
		if ( INDIR_A(mode) ) {
		    mode = RAW_MODE(mode);
		    pofs = &(pt->a);
		} else
		    pofs = &(pt->b);

		t = *pofs;      /* pofs is the indirection offset */
		if (mode == PREDEC) {
		    DECMOD(t);
		    *pofs = t;
		}
		ADDMOD(da, t, da);

		pt = core+da;       /* now pt is the final destination */
                /* of indiretion. */
		ra_a = pt->a;       /* read in registers */
		ra_b = pt->b;

		if (mode == POSTINC) {
		    INCMOD(t);
		    *pofs = t;
		}
	    } else /* DIRECT */{
		ra_a = pt->a;
		ra_b = pt->b;
	    }
	} else /* IMMEDIATE */ {
	    ra_b = in_b;
	    da = ip;
	}


	/*
	 * b-field effective address calculation
	 */
	mode = in & mMASK;
	in >>= mBITS;       /* shift out a-mode bits */

	if ( mode != IMMEDIATE ) {
	    ADDMOD( db, in_b, ip );
	    pt = core+db;

	    if ( mode != DIRECT ) {
		if ( INDIR_A(mode) ) {
		    mode = RAW_MODE(mode);
		    pofs = &(pt->a);
		} else
		    pofs = &(pt->b);

		t = *pofs;
		if (mode == PREDEC) {
		    DECMOD(t);
		    *pofs = t;
		}
		ADDMOD(db, t, db);

		pt = core+db;
		rb_a = pt->a;
		rb_b = pt->b;

		if (mode == POSTINC) {
		    INCMOD(t);
		    *pofs = t;
		}
	    } else /* DIRECT */{
		rb_a = pt->a;
		rb_b = pt->b;
	    }
	} else /* IMMEDIATE */ {
	    db = ip;
	}

#if DEBUG == 2
	/* Debug output */
	printf("%6d %4ld  %s  |%4ld, d %4ld,%4ld a %4ld,%4ld b %4ld,%4ld\n",
	       cycles, ip, debug_line,
	       w->nprocs, da, db, 
	       ra_a, ra_b, rb_a, rb_b );
#endif

	/*
	 * Execute the instruction on opcode.modifier
	 */
	switch ( in ) {

	case _OP(SPL, mA):
	case _OP(SPL, mB):
	case _OP(SPL, mAB):
	case _OP(SPL, mBA):
	case _OP(SPL, mX):
	case _OP(SPL, mF):
	case _OP(SPL, mI):
	    INCMOD(ip);
	    queue(ip);
	    if ( w->nprocs < Processes ) {
		++w->nprocs;
		queue(da);
	    }
	    goto noqueue;

	case _OP(DAT, mA):
	case _OP(DAT, mB):
	case _OP(DAT, mAB):
	case _OP(DAT, mBA):
	case _OP(DAT, mX):
	case _OP(DAT, mF):
	case _OP(DAT, mI):
	die:
	if ( --w->nprocs > 0 )
	    goto noqueue;
	w->pred->succ = w->succ;
	w->succ->pred = w->pred;
	*death_tab++ = w->id;
	cycles = cycles - cycles/alive_cnt; /* nC+k -> (n-1)C+k */
	if ( --alive_cnt <= 1 ) 
	    goto out;
	goto noqueue;

	case _OP(MOV, mA):
	    core[db].a = ra_a;
	    break;
	case _OP(MOV, mF):
	    core[db].a = ra_a;
	case _OP(MOV, mB):
	    core[db].b = ra_b;
	    break;
	case _OP(MOV, mAB):
	    core[db].b = ra_a;
	    break;
	case _OP(MOV, mX):
	    core[db].b = ra_a;
	case _OP(MOV, mBA):
	    core[db].a = ra_b;
	    break;
	case _OP(MOV, mI):
	    core[db].a = ra_a;
	    core[db].b = ra_b;
	    core[db].in = core[da].in;
	    break;

	case _OP(JMP, mA):
	case _OP(JMP, mB):
	case _OP(JMP, mAB):
	case _OP(JMP, mBA):
	case _OP(JMP, mX):
	case _OP(JMP, mF):
	case _OP(JMP, mI):
	    queue(da);
	    goto noqueue;

	case _OP(JMZ, mBA):
	case _OP(JMZ, mA):
	    if ( rb_a )
		break;
	    queue(da);
	    goto noqueue;

	case _OP(JMZ, mAB):
	case _OP(JMZ, mB):
	    if ( rb_b )
		break;
	    queue(da);
	    goto noqueue;

	case _OP(JMZ, mX):
	case _OP(JMZ, mF):
	case _OP(JMZ, mI):
	    if ( rb_a || rb_b )
		break;
	    queue(da);
	    goto noqueue;

	case _OP(JMN, mBA):
	case _OP(JMN, mA):
	    if (! rb_a )
		break;
	    queue(da);
	    goto noqueue;

	case _OP(JMN, mAB):
	case _OP(JMN, mB):
	    if (! rb_b )
		break;
	    queue(da);
	    goto noqueue;

	case _OP(JMN, mX):
	case _OP(JMN, mF):
	case _OP(JMN, mI):
	    if (!(rb_a | rb_b))
		break;
	    queue(da);
	    goto noqueue;


	case _OP(ADD, mI):
	case _OP(ADD, mF):
	    ADDMOD( t, ra_b, rb_b ); core[db].b = t;
	case _OP(ADD, mA):
	    ADDMOD( t, ra_a, rb_a ); core[db].a = t; 
	    break;
	case _OP(ADD, mB):
	    ADDMOD( t, ra_b, rb_b ); core[db].b = t;
	    break;
	case _OP(ADD, mX):
	    ADDMOD( t, ra_b, rb_a ); core[db].a = t;
	case _OP(ADD, mAB):
	    ADDMOD( t, ra_a, rb_b ); core[db].b = t;
	    break;
	case _OP(ADD, mBA):
	    ADDMOD( t, ra_b, rb_a ); core[db].a = t;
	    break;

	case _OP(SUB, mI):
	case _OP(SUB, mF):
	    SUBMOD( t, rb_b, ra_b ); core[db].b = t;
	case _OP(SUB, mA):
	    SUBMOD( t, rb_a, ra_a); core[db].a = t;
	    break;
	case _OP(SUB, mB):
	    SUBMOD( t, rb_b, ra_b ); core[db].b = t;
	    break;
	case _OP(SUB, mX):
	    SUBMOD( t, rb_a, ra_b ); core[db].a = t;
	case _OP(SUB, mAB):
	    SUBMOD( t, rb_b, ra_a ); core[db].b = t;
	    break;
	case _OP(SUB, mBA):
	    SUBMOD( t, rb_a, ra_b ); core[db].a = t;
	    break;

	case _OP(MUL, mI):
	case _OP(MUL, mF):
	    core[db].b = (rb_b * ra_b) % coresize;
	case _OP(MUL, mA):
	    core[db].a = (rb_a * ra_a) % coresize;
	    break;
	case _OP(MUL, mB):
	    core[db].b = (rb_b * ra_b) % coresize;
	    break;
	case _OP(MUL, mX):
	    core[db].a = (rb_a * ra_b) % coresize;
	case _OP(MUL, mAB):
	    core[db].b = (rb_b * ra_a) % coresize;
	    break;
	case _OP(MUL, mBA):
	    core[db].a = (rb_a * ra_b) % coresize;
	    break;

	case _OP(DIV, mI):
	case _OP(DIV, mF):
	    ftmp = 0;         /* we must try to do both divisions even if
				 one fails */
	    if ( ra_a ) core[db].a = rb_a / ra_a; else ftmp = 1;
	    if ( ra_b ) core[db].b = rb_b / ra_b; else ftmp = 1;
	    if ( ftmp ) goto die;
	    break;
	case _OP(DIV, mX):
	    ftmp = 0;         /* we must try to do both divisions even if
				 one fails */
	    if ( ra_b ) core[db].a = rb_a / ra_b; else ftmp = 1;
	    if ( ra_a ) core[db].b = rb_b / ra_a; else ftmp = 1;
	    if ( ftmp ) goto die;
	    break;
	case _OP(DIV, mA):
	    if ( !ra_a ) goto die;
	    core[db].a = rb_a / ra_a;
	    break;
	case _OP(DIV, mB):
	    if ( !ra_b ) goto die;
	    core[db].b = rb_b / ra_b;
	    break;
	case _OP(DIV, mAB):
	    if ( !ra_a ) goto die;
	    core[db].b = rb_b / ra_a;
	    break;
	case _OP(DIV, mBA):
	    if ( !ra_b ) goto die;
	    core[db].a = rb_a / ra_b;
	    break;

	case _OP(MODM, mI):
	case _OP(MODM, mF):
	    ftmp = 0;         /* we must try to do both modulos even if
				 one fails */
	    if ( ra_a ) core[db].a = rb_a % ra_a; else ftmp = 1;
	    if ( ra_b ) core[db].b = rb_b % ra_b; else ftmp = 1;
	    if ( ftmp ) goto die;
	    break;
	case _OP(MODM, mX):
	    ftmp = 0;         /* we must try to do both modulos even if
				 one fails */
	    if ( ra_b ) core[db].a = rb_a%ra_b; else ftmp = 1;
	    if ( ra_a ) core[db].b = rb_b%ra_a; else ftmp = 1;
	    if ( ftmp ) goto die;
	    break;
	case _OP(MODM, mA):
	    if ( !ra_a ) goto die;
	    core[db].a = rb_a % ra_a;
	    break;
	case _OP(MODM, mB):
	    if ( !ra_b ) goto die;
	    core[db].b = rb_b % ra_b;
	    break;
	case _OP(MODM, mAB):
	    if ( !ra_a ) goto die;
	    core[db].b = rb_b % ra_a;
	    break;
	case _OP(MODM, mBA):
	    if ( !ra_b ) goto die;
	    core[db].a = rb_a % ra_b;
	    break;


	case _OP(SEQ, mA):
	    if ( ra_a == rb_a )
		INCMOD(ip);
	    break;
	case _OP(SEQ, mB):
	    if ( ra_b == rb_b )
		INCMOD(ip);
	    break;
	case _OP(SEQ, mAB):
	    if ( ra_a == rb_b )
		INCMOD(ip);
	    break;
	case _OP(SEQ, mBA):
	    if ( ra_b == rb_a )
		INCMOD(ip);
	    break;

	case _OP(SEQ, mI):
	    if ( core[da].in != core[db].in )
		break;
	case _OP(SEQ, mF):
	    if (( (ra_a - rb_a) | (ra_b - rb_b)) == 0 )
		INCMOD(ip);
	    break;
	case _OP(SEQ, mX):
	    if ( ((ra_a - rb_b) | (ra_b - rb_a)) == 0 )
		INCMOD(ip);
	    break;

	case _OP(SNE, mA):
	    if ( ra_a != rb_a )
		INCMOD(ip);
	    break;
	case _OP(SNE, mB):
	    if ( ra_b != rb_b )
		INCMOD(ip);
	    break;
	case _OP(SNE, mAB):
	    if ( ra_a != rb_b )
		INCMOD(ip);
	    break;
	case _OP(SNE, mBA):
	    if ( ra_b != rb_a )
		INCMOD(ip);
	    break;

	case _OP(SNE, mI):
	    if ( core[da].in != core[db].in ) {
		INCMOD(ip);
		break;
	    }
	    /* fall through */
	case _OP(SNE, mF):
	    if ( ra_a != rb_a || ra_b != rb_b )
		INCMOD(ip);
	    break;
	case _OP(SNE, mX):
	    if ( ra_a != rb_b || ra_b != rb_a )
		INCMOD(ip);
	    break;

	case _OP(DJN,mBA):
	case _OP(DJN,mA):
	    t = core[db].a;
	    DECMOD(t);
	    core[db].a = t;
	    if ( rb_a == 1 ) break;
	    queue(da);
	    goto noqueue;

	case _OP(DJN,mAB):
	case _OP(DJN,mB):
	    t = core[db].b;
	    DECMOD(t);
	    core[db].b = t;
	    if ( rb_b == 1 ) break;
	    queue(da);
	    goto noqueue;

	case _OP(DJN,mX):
	case _OP(DJN,mI):
	case _OP(DJN,mF):
	    t = core[db].a;      DECMOD(t);      core[db].a = t;
	    t = core[db].b;      DECMOD(t);      core[db].b = t;
	    if ( ((rb_a -1) | (rb_b -1 ))==0) break;
	    queue(da);
	    goto noqueue;

	case _OP(SLT, mA):
	    if (ra_a < rb_a)
		INCMOD(ip);
	    break;
	case _OP(SLT, mAB):
	    if (ra_a < rb_b)
		INCMOD(ip);
	    break;
	case _OP(SLT, mB):
	    if (ra_b < rb_b)
		INCMOD(ip);
	    break;
	case _OP(SLT, mBA):
	    if (ra_b < rb_a)
		INCMOD(ip);
	    break;
	case _OP(SLT, mI):
	case _OP(SLT, mF):
	    if (ra_a < rb_a && ra_b < rb_b)
		INCMOD(ip);
	    break;
	case _OP(SLT, mX):
	    if (ra_a < rb_b && ra_b < rb_a)
		INCMOD(ip);
	    break;

	case _OP(NOP,mI):
	case _OP(NOP,mX):
	case _OP(NOP,mF):
	case _OP(NOP,mA):
	case _OP(NOP,mAB):
	case _OP(NOP,mB):
	case _OP(NOP,mBA):
	    break;

	default:
	    return -1;
	}

	INCMOD(ip);
	queue(ip);
    noqueue:
	w = w->succ;
    } while(--cycles>0);

out:
    return alive_cnt;
}
