/* This file is part of 'yace', a corewar-warrior evolver.
 *
 * Copyright (C) 2000 Martin Ankerl <martinankerl@web.de>
 *    
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "utils/exhaust/all.h"
#include "utils/utils.h"
#include "fitness/fitness.h"
#include "select/select.h"
#include "replace/replace.h"
#include "scoresel/scoresel.h"


int main(int argc, char* argv[]) {
    /* w is a pointer to the first element of some warriors */
    warrior_t * w=0;
    warrior_select_t * wsel=0;
    long current_cycle;
    int fitness;
    

    /* do a benchmark */
    if (argc >= 2) {
	if(strstr(argv[1], "bench")) {
	    benchmark();
	    exit(0);
	}
	else Intro();
	exit(0);
    }

    /* load config file settings, allocate memory, creates del2stop-file */
    setup(&w, &wsel);

    /* if warriors exist in directory "soup" load them, otherwise create them */
    create_warriors(w, globals.population_size, globals.startlength, "soup");
    
    /* cycle util file 'del2stop' is deleted (tested in shutdowntest()) */
    for(current_cycle=0; !shutdowntest(); current_cycle++) {
        printf("cycle %5ld\n",current_cycle);

        
        /*************************************************************/
        /* calculate scores for all warriors                         */
        /*************************************************************/
        fitness = globals.evolver_fitness;
        if (fitness == 100) fitness = random_number(5);
                
        switch(fitness) {
            
            /* selection + fitness as in yace 0.02, similar to ga_war.c */
        case 0: 
            select_standard(w, wsel, globals.population_size);
            fitness_standard(wsel);
            break;

            /* evolve with a hill! creates status.htm in directory 'koth'*/
        case 1:
            /* no selection needed. last parameter=0: no warrior will
               enter the hill, only scores are calculated (unprecise) */
            fitness_koth(w, globals.population_size, 0);
            /* see if best warrior can enter the hill. last
               parameter=1: if warrior is good enough it will enter,
               this is calculated precise */
            fitness_koth(&w[get_best_warrior(w, globals.population_size)], 1, 1);
            break;
            
            /* multiwar: every warrior fights 1 battle against some warriors */
        case 2:
            select_standard(w, wsel, globals.population_size);
            fitness_multiwar(w, globals.population_size);
            break;

            /* evolve with a hill, replace most similar warrior if better */
        case 3:
            /* no selection needed. last parameter=0: no warrior will
               enter the hill, only scores are calculated (unprecise) */
            fitness_kothsimi(w, globals.population_size, 0);
            /* see if best warrior can enter the hill. last
               parameter=1: if warrior is good enough it will enter,
               this is calculated precise */
            fitness_kothsimi(&w[get_best_warrior(w, globals.population_size)], 1, 1);
            break;

            
            /* testset: fight against testset */
        case 4:
            fitness_testset(w, globals.population_size, 0);
            break;            
        }
        

        /*************************************************************/
        /* select warriors for reproduction                          */
        /*************************************************************/
        switch (globals.evolver_replacement) {

            /* the best 1/4 of the soup is selected to replace the bad 1/2
               of the soup */
        case 0:
            scoresel_weighted(w, wsel, globals.population_size);
            break;

            /* WARNING! only works if globals.evolver_fitness==0 !! */
        case 1:
            if (globals.evolver_fitness == 0) scoresel_standard(wsel);
            else scoresel_weighted(w, wsel, globals.population_size);
            break;
        }        
        
        /*************************************************************/
        /* replace warriors                                          */
        /*************************************************************/
        /* this is the only replacement routine so far, you are
           welcome to write your own one :-) */
        replace_standard(wsel);
        

        /*************************************************************/
        /* administrative tasks                                      */
        /*************************************************************/     
        /* Save population to disk */
        if (globals.cycles_before_save && !(current_cycle % globals.cycles_before_save) && current_cycle)
            save_warriors(w, globals.population_size, current_cycle, "soup");
        
        /* Status update: extract best warrior */
        if (globals.cycles_before_status  && !(current_cycle % globals.cycles_before_status) && current_cycle) {
            status(w, globals.population_size, current_cycle);
            /* status usese fitness_testset, so we can reproduce with this results */
            scoresel_weighted(w, wsel, globals.population_size);
            replace_standard(wsel);
        }
    }
    printf("Shutting down...\n");
    save_warriors(w, globals.population_size, current_cycle, "soup");
    printf("%10ld fights\n", globals.all_battles);
    return 0;
}
